/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
/* eslint-disable @typescript-eslint/no-non-null-assertion */
import { TypedEventEmitter } from "@fluid-internal/client-utils";
import { assert, unreachableCase, isObject } from "@fluidframework/core-utils/internal";
import { MessageType, } from "@fluidframework/driver-definitions/internal";
import { toDeltaManagerInternal } from "@fluidframework/runtime-utils/internal";
import { LoggingError, UsageError, } from "@fluidframework/telemetry-utils/internal";
import { MergeTreeTextHelper } from "./MergeTreeTextHelper.js";
import { RedBlackTree } from "./collections/index.js";
import { UnassignedSequenceNumber, UniversalSequenceNumber } from "./constants.js";
import { MergeTree, errorIfOptionNotTrue } from "./mergeTree.js";
import { walkAllChildSegments } from "./mergeTreeNodeWalk.js";
import { compareStrings, toMoveInfo, } from "./mergeTreeNodes.js";
import { createAnnotateMarkerOp, createAnnotateRangeOp, 
// eslint-disable-next-line import/no-deprecated
createGroupOp, createInsertSegmentOp, createObliterateRangeOp, createObliterateRangeOpSided, createRemoveRangeOp, } from "./opBuilder.js";
import { MergeTreeDeltaType, } from "./ops.js";
import { DetachedReferencePosition } from "./referencePositions.js";
import { Side } from "./sequencePlace.js";
import { SnapshotLoader } from "./snapshotLoader.js";
import { SnapshotV1 } from "./snapshotV1.js";
import { SnapshotLegacy } from "./snapshotlegacy.js";
/**
 * @deprecated This functionality was not meant to be exported and will be removed in a future release
 * @legacy
 * @alpha
 */
export class Client extends TypedEventEmitter {
    /**
     * @param specToSegment - Rehydrates a segment from its JSON representation
     * @param logger - Telemetry logger for diagnostics
     * @param options - Options for this client. See {@link IMergeTreeOptions} for details.
     * @param getMinInFlightRefSeq - Upon applying a message (see {@link Client.applyMsg}), client purges collab-window information which
     * is no longer necessary based on that message's minimum sequence number.
     * However, if the user of this client has in-flight messages which refer to positions in this Client,
     * they may wish to preserve additional merge information.
     * The effective minimum sequence number will be the minimum of the message's minimumSequenceNumber and the result of this function.
     * If this function returns undefined, the message's minimumSequenceNumber will be used.
     *
     * @privateRemarks
     * - Passing specToSegment would be unnecessary if Client were merged with SharedSegmentSequence
     * - AB#6866 tracks a more unified approach to collab window min seq handling.
     */
    constructor(specToSegment, logger, options, getMinInFlightRefSeq = () => undefined) {
        super();
        this.specToSegment = specToSegment;
        this.logger = logger;
        this.getMinInFlightRefSeq = getMinInFlightRefSeq;
        this.clientNameToIds = new RedBlackTree(compareStrings);
        this.shortClientIdMap = [];
        this.lastNormalizationRefSeq = 0;
        this._mergeTree = new MergeTree(options);
        this._mergeTree.mergeTreeDeltaCallback = (opArgs, deltaArgs) => {
            this.emit("delta", opArgs, deltaArgs, this);
        };
        this._mergeTree.mergeTreeMaintenanceCallback = (args, opArgs) => {
            this.emit("maintenance", args, opArgs, this);
        };
        if (options?.attribution?.track) {
            const policy = this._mergeTree?.attributionPolicy;
            if (policy === undefined) {
                throw new UsageError("Attribution policy must be provided when attribution tracking is requested.");
            }
            policy.attach(this);
        }
    }
    peekPendingSegmentGroups(count = 1) {
        const pending = this._mergeTree.pendingSegments;
        let node = pending?.last;
        if (count === 1 || pending === undefined) {
            return node?.data;
        }
        // eslint-disable-next-line import/no-deprecated
        const taken = Array.from({ length: Math.min(count, pending.length) });
        for (let i = taken.length - 1; i >= 0; i--) {
            taken[i] = node.data;
            node = node.prev;
        }
        return taken;
    }
    /**
     * Annotates the markers with the provided properties
     * @param marker - The marker to annotate
     * @param props - The properties to annotate the marker with
     * @returns The annotate op if valid, otherwise undefined
     */
    annotateMarker(marker, props) {
        const annotateOp = createAnnotateMarkerOp(marker, props);
        this.applyAnnotateRangeOp({ op: annotateOp });
        return annotateOp;
    }
    /**
     * Annotates the range with the provided properties
     * @param start - The inclusive start position of the range to annotate
     * @param end - The exclusive end position of the range to annotate
     * @param props - The properties to annotate the range with
     * @returns The annotate op if valid, otherwise undefined
     */
    annotateRangeLocal(start, end, props) {
        const annotateOp = createAnnotateRangeOp(start, end, props);
        this.applyAnnotateRangeOp({ op: annotateOp });
        return annotateOp;
    }
    /**
     * Removes the range
     *
     * @param start - The inclusive start of the range to remove
     * @param end - The exclusive end of the range to remove
     */
    removeRangeLocal(start, end) {
        const removeOp = createRemoveRangeOp(start, end);
        this.applyRemoveRangeOp({ op: removeOp });
        return removeOp;
    }
    /**
     * Obliterates the range. This is similar to removing the range, but also
     * includes any concurrently inserted content.
     *
     * @param start - The start of the range to obliterate. Inclusive is side is Before (default).
     * @param end - The end of the range to obliterate. Exclusive is side is After
     * (default is to be after the last included character, but number index is exclusive).
     */
    obliterateRangeLocal(start, end) {
        // eslint-disable-next-line import/no-deprecated
        let obliterateOp;
        if (this._mergeTree.options?.mergeTreeEnableSidedObliterate) {
            obliterateOp = createObliterateRangeOpSided(start, end);
        }
        else {
            assert(typeof start === "number" && typeof end === "number", "Start and end must be numbers if mergeTreeEnableSidedObliterate is not enabled.");
            obliterateOp = createObliterateRangeOp(start, end);
        }
        this.applyObliterateRangeOp({ op: obliterateOp });
        return obliterateOp;
    }
    /**
     * Create and insert a segment at the specified position.
     * @param pos - The position to insert the segment at
     * @param segment - The segment to insert
     */
    insertSegmentLocal(pos, segment) {
        if (segment.cachedLength <= 0) {
            return undefined;
        }
        const insertOp = createInsertSegmentOp(pos, segment);
        this.applyInsertOp({ op: insertOp });
        return insertOp;
    }
    /**
     * Create and insert a segment at the specified reference position.
     * @param refPos - The reference position to insert the segment at
     * @param segment - The segment to insert
     */
    insertAtReferencePositionLocal(refPos, segment) {
        const pos = this._mergeTree.referencePositionToLocalPosition(refPos, this.getCurrentSeq(), this.getClientId());
        if (pos === DetachedReferencePosition) {
            throw new UsageError("Cannot insert at detached local reference.");
        }
        return this.insertSegmentLocal(pos, segment);
    }
    walkSegments(handler, start, end, accum, splitRange = false) {
        this._mergeTree.mapRange(handler, this.getCurrentSeq(), this.getClientId(), accum, start, end, splitRange);
    }
    walkAllSegments(action, accum) {
        return walkAllChildSegments(this._mergeTree.root, accum === undefined ? action : (seg) => action(seg, accum));
    }
    /**
     * Serializes the data required for garbage collection. The IFluidHandles stored in all segments that haven't
     * been removed represent routes to other objects. We serialize the data in these segments using the passed in
     * serializer which keeps track of all serialized handles.
     */
    serializeGCData(handle, handleCollectingSerializer) {
        let localInserts = 0;
        let localRemoves = 0;
        walkAllChildSegments(this._mergeTree.root, (seg) => {
            if (seg.seq === UnassignedSequenceNumber) {
                localInserts++;
            }
            if (seg.removedSeq === UnassignedSequenceNumber) {
                localRemoves++;
            }
            // Only serialize segments that have not been removed.
            if (seg.removedSeq === undefined) {
                handleCollectingSerializer.stringify(seg.clone().toJSONObject(), handle);
            }
            return true;
        });
        if (localInserts > 0 || localRemoves > 0) {
            this.logger.sendErrorEvent({
                eventName: "LocalEditsInProcessGCData",
                localInserts,
                localRemoves,
            });
        }
    }
    // eslint-disable-next-line import/no-deprecated
    getCollabWindow() {
        return this._mergeTree.collabWindow;
    }
    /**
     * Returns the current position of a segment, and -1 if the segment
     * does not exist in this merge tree
     * @param segment - The segment to get the position of
     */
    getPosition(segment, localSeq) {
        const mergeSegment = segment;
        if (mergeSegment?.parent === undefined) {
            return -1;
        }
        return this._mergeTree.getPosition(mergeSegment, this.getCurrentSeq(), this.getClientId(), localSeq);
    }
    /**
     * Creates a `LocalReferencePosition` on this client. If the refType does not include ReferenceType.Transient,
     * the returned reference will be added to the localRefs on the provided segment.
     * @param segment - Segment to add the local reference on
     * @param offset - Offset on the segment at which to place the local reference
     * @param refType - ReferenceType for the created local reference
     * @param properties - PropertySet to place on the created local reference
     * @param canSlideToEndpoint - Whether or not the created local reference can
     * slide onto one of the special endpoint segments denoting the position
     * before the start of or after the end of the tree
     */
    createLocalReferencePosition(segment, offset, refType, properties, slidingPreference, canSlideToEndpoint) {
        return this._mergeTree.createLocalReferencePosition(segment, offset ?? 0, refType, properties, slidingPreference, canSlideToEndpoint);
    }
    /**
     * Removes a `LocalReferencePosition` from this client.
     */
    removeLocalReferencePosition(lref) {
        return this._mergeTree.removeLocalReferencePosition(lref);
    }
    /**
     * Resolves a `ReferencePosition` into a character position using this client's perspective.
     *
     * Reference positions that point to a character that has been removed will
     * always return the position of the nearest non-removed character, regardless
     * of {@link ReferenceType}. To handle this case specifically, one may wish
     * to look at the segment returned by {@link ReferencePosition.getSegment}.
     */
    localReferencePositionToPosition(lref) {
        return this._mergeTree.referencePositionToLocalPosition(lref);
    }
    /**
     * Given a position specified relative to a marker id, lookup the marker
     * and convert the position to a character position.
     * @param relativePos - Id of marker (may be indirect) and whether position is before or after marker.
     */
    posFromRelativePos(relativePos) {
        return this._mergeTree.posFromRelativePos(relativePos);
    }
    getMarkerFromId(id) {
        return this._mergeTree.getMarkerFromId(id);
    }
    /**
     * Revert an op
     */
    rollback(op, localOpMetadata) {
        // eslint-disable-next-line import/no-deprecated
        this._mergeTree.rollback(op, localOpMetadata);
    }
    applyObliterateRangeOp(opArgs) {
        assert(opArgs.op.type === MergeTreeDeltaType.OBLITERATE ||
            opArgs.op.type === MergeTreeDeltaType.OBLITERATE_SIDED, 0x866 /* Unexpected op type on range obliterate! */);
        const op = opArgs.op;
        const clientArgs = this.getClientSequenceArgs(opArgs);
        if (this._mergeTree.options?.mergeTreeEnableSidedObliterate) {
            const { start, end } = this.getValidSidedRange(op, clientArgs);
            this._mergeTree.obliterateRange(start, end, clientArgs.referenceSequenceNumber, clientArgs.clientId, clientArgs.sequenceNumber, opArgs);
        }
        else {
            assert(op.type === MergeTreeDeltaType.OBLITERATE, "Unexpected sided obliterate while mergeTreeEnableSidedObliterate is disabled");
            const range = this.getValidOpRange(op, clientArgs);
            this._mergeTree.obliterateRange(range.start, range.end, clientArgs.referenceSequenceNumber, clientArgs.clientId, clientArgs.sequenceNumber, opArgs);
        }
    }
    /**
     * Performs the remove based on the provided op
     * @param opArgs - The ops args for the op
     */
    applyRemoveRangeOp(opArgs) {
        assert(opArgs.op.type === MergeTreeDeltaType.REMOVE, 0x02d /* "Unexpected op type on range remove!" */);
        const op = opArgs.op;
        const clientArgs = this.getClientSequenceArgs(opArgs);
        const range = this.getValidOpRange(op, clientArgs);
        this._mergeTree.markRangeRemoved(range.start, range.end, clientArgs.referenceSequenceNumber, clientArgs.clientId, clientArgs.sequenceNumber, opArgs);
    }
    /**
     * Performs the annotate based on the provided op
     * @param opArgs - The ops args for the op
     */
    applyAnnotateRangeOp(opArgs) {
        assert(opArgs.op.type === MergeTreeDeltaType.ANNOTATE, 0x02e /* "Unexpected op type on range annotate!" */);
        const op = opArgs.op;
        const clientArgs = this.getClientSequenceArgs(opArgs);
        const range = this.getValidOpRange(op, clientArgs);
        this._mergeTree.annotateRange(range.start, range.end, op.props, clientArgs.referenceSequenceNumber, clientArgs.clientId, clientArgs.sequenceNumber, opArgs);
    }
    /**
     * Performs the insert based on the provided op
     * @param opArgs - The ops args for the op
     * @returns True if the insert was applied. False if it could not be.
     */
    applyInsertOp(opArgs) {
        assert(opArgs.op.type === MergeTreeDeltaType.INSERT, 0x02f /* "Unexpected op type on range insert!" */);
        const op = opArgs.op;
        const clientArgs = this.getClientSequenceArgs(opArgs);
        const range = this.getValidOpRange(op, clientArgs);
        // eslint-disable-next-line @typescript-eslint/no-unsafe-argument
        const segments = [this.specToSegment(op.seg)];
        this._mergeTree.insertSegments(range.start, segments, clientArgs.referenceSequenceNumber, clientArgs.clientId, clientArgs.sequenceNumber, opArgs);
    }
    /**
     * Returns a valid range for the op, or throws if the range is invalid
     * @param op - The op to generate the range for
     * @param clientArgs - The client args for the op
     * @throws LoggingError if the range is invalid
     */
    getValidSidedRange(
    // eslint-disable-next-line import/no-deprecated
    op, clientArgs) {
        const invalidPositions = [];
        let start;
        let end;
        if (op.pos1 === undefined) {
            invalidPositions.push("start");
        }
        else {
            start =
                typeof op.pos1 === "object"
                    ? { pos: op.pos1.pos, side: op.pos1.before ? Side.Before : Side.After }
                    : { pos: op.pos1, side: Side.Before };
        }
        if (op.pos2 === undefined) {
            invalidPositions.push("end");
        }
        else {
            end =
                typeof op.pos2 === "object"
                    ? { pos: op.pos2.pos, side: op.pos2.before ? Side.Before : Side.After }
                    : { pos: op.pos2 - 1, side: Side.After };
        }
        // Validate if local op
        if (clientArgs.clientId === this.getClientId()) {
            const length = this._mergeTree.getLength(this.getCollabWindow().currentSeq, this.getClientId());
            if (start !== undefined && (start.pos >= length || start.pos < 0)) {
                // start out of bounds
                invalidPositions.push("start");
            }
            if (end !== undefined && (end.pos >= length || end.pos < 0)) {
                invalidPositions.push("end");
            }
            if (start !== undefined &&
                end !== undefined &&
                (start.pos > end.pos ||
                    (start.pos === end.pos && start.side !== end.side && start.side === Side.After))) {
                // end is before start
                invalidPositions.push("inverted");
            }
            if (invalidPositions.length > 0) {
                throw new LoggingError("InvalidRange", {
                    usageError: true,
                    invalidPositions: invalidPositions.toString(),
                    length,
                    opType: op.type,
                    opPos1Relative: op.relativePos1 !== undefined,
                    opPos2Relative: op.relativePos2 !== undefined,
                    opPos1: JSON.stringify(op.pos1),
                    opPos2: JSON.stringify(op.pos2),
                    start: JSON.stringify(start),
                    end: JSON.stringify(end),
                });
            }
        }
        assert(start !== undefined && end !== undefined, "Missing start or end of range");
        return { start, end };
    }
    /**
     * Returns a valid range for the op, or undefined
     * @param op - The op to generate the range for
     * @param clientArgs - The client args for the op
     */
    getValidOpRange(op, clientArgs) {
        let start = op.pos1;
        if (start === undefined && op.relativePos1) {
            start = this._mergeTree.posFromRelativePos(op.relativePos1, clientArgs.referenceSequenceNumber, clientArgs.clientId);
        }
        let end = op.pos2;
        if (end === undefined && op.relativePos2) {
            end = this._mergeTree.posFromRelativePos(op.relativePos2, clientArgs.referenceSequenceNumber, clientArgs.clientId);
        }
        // Validate if local op
        if (clientArgs.clientId === this.getClientId()) {
            const length = this.getLength();
            const invalidPositions = [];
            // Validate start position
            //
            if (start === undefined ||
                start < 0 ||
                start > length ||
                (start === length && op.type !== MergeTreeDeltaType.INSERT)) {
                invalidPositions.push("start");
            }
            // Validate end if not insert, or insert has end
            if ((op.type !== MergeTreeDeltaType.INSERT || end !== undefined) &&
                (end === undefined || end <= start)) {
                invalidPositions.push("end");
            }
            if (op.type === MergeTreeDeltaType.OBLITERATE && end !== undefined && end > length) {
                invalidPositions.push("end");
            }
            if (invalidPositions.length > 0) {
                throw new LoggingError("RangeOutOfBounds", {
                    usageError: true,
                    end,
                    invalidPositions: invalidPositions.toString(),
                    length,
                    opPos1: op.pos1,
                    opPos1Relative: op.relativePos1 !== undefined,
                    opPos2: op.pos2,
                    opPos2Relative: op.relativePos2 !== undefined,
                    opType: op.type,
                    start,
                });
            }
        }
        // start and end are guaranteed to be non-null here, otherwise we throw above.
        return { start: start, end: end };
    }
    /**
     * Gets the client args from the op if remote, otherwise uses the local clients info
     * @param sequencedMessage - The sequencedMessage to get the client sequence args for
     */
    getClientSequenceArgsForMessage(sequencedMessage) {
        // If there this no sequenced message, then the op is local
        // and unacked, so use this clients sequenced args
        //
        if (sequencedMessage) {
            return {
                clientId: this.getOrAddShortClientIdFromMessage(sequencedMessage),
                referenceSequenceNumber: sequencedMessage.referenceSequenceNumber,
                // Note: return value satisfies overload signatures despite the cast, as if input argument doesn't contain sequenceNumber,
                // return value isn't expected to have it either.
                sequenceNumber: sequencedMessage.sequenceNumber,
            };
        }
        else {
            const segWindow = this.getCollabWindow();
            return {
                clientId: segWindow.clientId,
                referenceSequenceNumber: segWindow.currentSeq,
                sequenceNumber: this.getLocalSequenceNumber(),
            };
        }
    }
    /**
     * Gets the client args from the op if remote, otherwise uses the local clients info
     * @param opArgs - The op arg to get the client sequence args for
     */
    getClientSequenceArgs(opArgs) {
        return this.getClientSequenceArgsForMessage(opArgs.sequencedMessage);
    }
    ackPendingSegment(opArgs) {
        if (opArgs.op.type === MergeTreeDeltaType.GROUP) {
            for (const memberOp of opArgs.op.ops) {
                this._mergeTree.ackPendingSegment({
                    groupOp: opArgs.op,
                    op: memberOp,
                    sequencedMessage: opArgs.sequencedMessage,
                });
            }
        }
        else {
            this._mergeTree.ackPendingSegment(opArgs);
        }
    }
    getOrAddShortClientId(longClientId) {
        if (!this.clientNameToIds.get(longClientId)) {
            this.addLongClientId(longClientId);
        }
        return this.getShortClientId(longClientId);
    }
    getShortClientId(longClientId) {
        return this.clientNameToIds.get(longClientId).data;
    }
    getLongClientId(shortClientId) {
        return shortClientId >= 0 ? this.shortClientIdMap[shortClientId] : "original";
    }
    addLongClientId(longClientId) {
        this.clientNameToIds.put(longClientId, this.shortClientIdMap.length);
        this.shortClientIdMap.push(longClientId);
    }
    getOrAddShortClientIdFromMessage(msg) {
        return this.getOrAddShortClientId(msg.clientId ?? "server");
    }
    /**
     * During reconnect, we must find the positions to pending segments
     * relative to other pending segments. This methods computes that
     * position relative to a localSeq. Pending segments above the localSeq
     * will be ignored.
     *
     * @param segment - The segment to find the position for
     * @param localSeq - The localSeq to find the position of the segment at
     */
    findReconnectionPosition(segment, localSeq) {
        assert(localSeq <= this._mergeTree.collabWindow.localSeq, 0x032 /* "localSeq greater than collab window" */);
        const { currentSeq, clientId } = this.getCollabWindow();
        return this._mergeTree.getPosition(segment, currentSeq, clientId, localSeq);
    }
    resetPendingDeltaToOps(resetOp, 
    // eslint-disable-next-line import/no-deprecated
    segmentGroup) {
        assert(!!segmentGroup, 0x033 /* "Segment group undefined" */);
        const NACKedSegmentGroup = this.pendingRebase?.shift()?.data;
        assert(segmentGroup === NACKedSegmentGroup, 0x034 /* "Segment group not at head of pending rebase queue" */);
        if (this.pendingRebase?.empty) {
            this.pendingRebase = undefined;
        }
        // if this is an obliterate op, keep all segments in same segment group
        // eslint-disable-next-line import/no-deprecated
        const obliterateSegmentGroup = {
            segments: [],
            localSeq: segmentGroup.localSeq,
            refSeq: this.getCollabWindow().currentSeq,
            obliterateInfo: {
                ...segmentGroup.obliterateInfo,
            },
        };
        const opList = [];
        // We need to sort the segments by ordinal, as the segments are not sorted in the segment group.
        // The reason they need them sorted, as they have the same local sequence number and which means
        // farther segments will  take into account nearer segments when calculating their position.
        // By sorting we ensure the nearer segment will be applied and sequenced before the farther segments
        // so their recalculated positions will be correct.
        for (const segment of segmentGroup.segments.sort((a, b) => a.ordinal < b.ordinal ? -1 : 1)) {
            assert(segment.segmentGroups.remove?.(segmentGroup) === true, 0x035 /* "Segment group not in segment pending queue" */);
            assert(segmentGroup.localSeq !== undefined, 0x867 /* expected segment group localSeq to be defined */);
            const segmentPosition = this.findReconnectionPosition(segment, segmentGroup.localSeq);
            let newOp;
            switch (resetOp.type) {
                case MergeTreeDeltaType.ANNOTATE: {
                    assert(segment.propertyManager?.hasPendingProperties(resetOp.props) === true, 0x036 /* "Segment has no pending properties" */);
                    // if the segment has been removed or obliterated, there's no need to send the annotate op
                    // unless the remove was local, in which case the annotate must have come
                    // before the remove
                    if ((segment.removedSeq === undefined ||
                        (segment.localRemovedSeq !== undefined &&
                            segment.removedSeq === UnassignedSequenceNumber)) &&
                        (segment.movedSeq === undefined ||
                            (segment.localMovedSeq !== undefined &&
                                segment.movedSeq === UnassignedSequenceNumber))) {
                        newOp = createAnnotateRangeOp(segmentPosition, segmentPosition + segment.cachedLength, resetOp.props);
                    }
                    break;
                }
                case MergeTreeDeltaType.INSERT: {
                    assert(segment.seq === UnassignedSequenceNumber, 0x037 /* "Segment already has assigned sequence number" */);
                    const moveInfo = toMoveInfo(segment);
                    if (moveInfo !== undefined) {
                        errorIfOptionNotTrue(this._mergeTree.options, "mergeTreeEnableObliterateReconnect");
                        if (moveInfo.movedSeq !== UnassignedSequenceNumber) {
                            // the segment was remotely obliterated, so is considered removed
                            // we set the seq to the universal seq and remove the local seq,
                            // so its length is not considered for subsequent local changes
                            // this allows us to not send the op as even the local client will ignore the segment
                            segment.seq = UniversalSequenceNumber;
                            segment.localSeq = undefined;
                            break;
                        }
                    }
                    const segInsertOp = segment.clone();
                    const opProps = isObject(resetOp.seg) && "props" in resetOp.seg && isObject(resetOp.seg.props)
                        ? { ...resetOp.seg.props }
                        : undefined;
                    segInsertOp.properties = opProps;
                    newOp = createInsertSegmentOp(segmentPosition, segInsertOp);
                    break;
                }
                case MergeTreeDeltaType.REMOVE: {
                    if (segment.localRemovedSeq !== undefined &&
                        segment.removedSeq === UnassignedSequenceNumber &&
                        (segment.movedSeq === undefined ||
                            (segment.localMovedSeq !== undefined &&
                                segment.movedSeq === UnassignedSequenceNumber))) {
                        newOp = createRemoveRangeOp(segmentPosition, segmentPosition + segment.cachedLength);
                    }
                    break;
                }
                case MergeTreeDeltaType.OBLITERATE: {
                    errorIfOptionNotTrue(this._mergeTree.options, "mergeTreeEnableObliterateReconnect");
                    if (segment.localMovedSeq !== undefined &&
                        segment.movedSeq === UnassignedSequenceNumber &&
                        (segment.removedSeq === undefined ||
                            (segment.localRemovedSeq !== undefined &&
                                segment.removedSeq === UnassignedSequenceNumber))) {
                        newOp = createObliterateRangeOp(segmentPosition, segmentPosition + segment.cachedLength);
                    }
                    break;
                }
                default: {
                    throw new Error(`Invalid op type`);
                }
            }
            if (newOp && resetOp.type === MergeTreeDeltaType.OBLITERATE) {
                segment.segmentGroups.enqueue(obliterateSegmentGroup);
                const first = opList[0];
                if (!!first &&
                    first.pos2 !== undefined &&
                    first.type !== MergeTreeDeltaType.OBLITERATE_SIDED &&
                    newOp.type !== MergeTreeDeltaType.OBLITERATE_SIDED) {
                    first.pos2 += newOp.pos2 - newOp.pos1;
                }
                else {
                    opList.push(newOp);
                }
            }
            else if (newOp) {
                // eslint-disable-next-line import/no-deprecated
                const newSegmentGroup = {
                    segments: [],
                    localSeq: segmentGroup.localSeq,
                    refSeq: this.getCollabWindow().currentSeq,
                };
                segment.segmentGroups.enqueue(newSegmentGroup);
                this._mergeTree.pendingSegments.push(newSegmentGroup);
                opList.push(newOp);
            }
        }
        if (resetOp.type === MergeTreeDeltaType.OBLITERATE &&
            obliterateSegmentGroup.segments.length > 0) {
            this._mergeTree.pendingSegments.push(obliterateSegmentGroup);
        }
        return opList;
    }
    applyRemoteOp(opArgs) {
        const op = opArgs.op;
        const msg = opArgs.sequencedMessage;
        this.getOrAddShortClientIdFromMessage(msg);
        switch (op.type) {
            case MergeTreeDeltaType.INSERT: {
                this.applyInsertOp(opArgs);
                break;
            }
            case MergeTreeDeltaType.REMOVE: {
                this.applyRemoveRangeOp(opArgs);
                break;
            }
            case MergeTreeDeltaType.ANNOTATE: {
                this.applyAnnotateRangeOp(opArgs);
                break;
            }
            case MergeTreeDeltaType.OBLITERATE:
            case MergeTreeDeltaType.OBLITERATE_SIDED: {
                this.applyObliterateRangeOp(opArgs);
                break;
            }
            case MergeTreeDeltaType.GROUP: {
                for (const memberOp of op.ops) {
                    this.applyRemoteOp({
                        op: memberOp,
                        groupOp: op,
                        sequencedMessage: msg,
                    });
                }
                break;
            }
            default: {
                break;
            }
        }
    }
    applyStashedOp(op) {
        switch (op.type) {
            case MergeTreeDeltaType.INSERT: {
                this.applyInsertOp({ op });
                break;
            }
            case MergeTreeDeltaType.REMOVE: {
                this.applyRemoveRangeOp({ op });
                break;
            }
            case MergeTreeDeltaType.ANNOTATE: {
                this.applyAnnotateRangeOp({ op });
                break;
            }
            case MergeTreeDeltaType.OBLITERATE_SIDED:
            case MergeTreeDeltaType.OBLITERATE: {
                this.applyObliterateRangeOp({ op });
                break;
            }
            case MergeTreeDeltaType.GROUP: {
                op.ops.map((o) => this.applyStashedOp(o));
                break;
            }
            default: {
                unreachableCase(op, "unrecognized op type");
            }
        }
    }
    applyMsg(msg, local = false) {
        // Ensure client ID is registered
        this.getOrAddShortClientIdFromMessage(msg);
        // Apply if an operation message
        if (msg.type === MessageType.Operation) {
            const opArgs = {
                op: msg.contents,
                sequencedMessage: msg,
            };
            if (opArgs.sequencedMessage?.clientId === this.longClientId || local) {
                this.ackPendingSegment(opArgs);
            }
            else {
                this.applyRemoteOp(opArgs);
            }
        }
        const min = Math.min(this.getMinInFlightRefSeq() ?? Number.POSITIVE_INFINITY, msg.minimumSequenceNumber);
        this.updateSeqNumbers(min, msg.sequenceNumber);
    }
    updateSeqNumbers(min, seq) {
        const collabWindow = this.getCollabWindow();
        // Equal is fine here due to SharedSegmentSequence<>.snapshotContent() potentially updating with same #
        assert(collabWindow.currentSeq <= seq, 0x038 /* "Incoming op sequence# < local collabWindow's currentSequence#" */);
        collabWindow.currentSeq = seq;
        assert(min <= seq, 0x039 /* "Incoming op sequence# < minSequence#" */);
        this.updateMinSeq(min);
    }
    /**
     * Resolves a remote client's position against the local sequence
     * and returns the remote client's position relative to the local
     * sequence
     * @param remoteClientPosition - The remote client's position to resolve
     * @param remoteClientRefSeq - The reference sequence number of the remote client
     * @param remoteClientId - The client id of the remote client
     */
    resolveRemoteClientPosition(remoteClientPosition, remoteClientRefSeq, remoteClientId) {
        const shortRemoteClientId = this.getOrAddShortClientId(remoteClientId);
        return this._mergeTree.resolveRemoteClientPosition(remoteClientPosition, remoteClientRefSeq, shortRemoteClientId);
    }
    /**
     * Given a pending operation and segment group, regenerate the op, so it
     * can be resubmitted
     * @param resetOp - The op to reset
     * @param segmentGroup - The segment group associated with the op
     */
    regeneratePendingOp(resetOp, 
    // eslint-disable-next-line import/no-deprecated
    segmentGroup) {
        if (this.pendingRebase === undefined || this.pendingRebase.empty) {
            // eslint-disable-next-line import/no-deprecated
            let firstGroup;
            if (Array.isArray(segmentGroup)) {
                if (segmentGroup.length === 0) {
                    // sometimes we rebase to an empty op
                    // eslint-disable-next-line import/no-deprecated
                    return createGroupOp();
                }
                firstGroup = segmentGroup[0];
            }
            else {
                firstGroup = segmentGroup;
            }
            const firstGroupNode = this._mergeTree.pendingSegments.find((node) => node.data === firstGroup);
            assert(firstGroupNode !== undefined, 0x70e /* segment group must exist in pending list */);
            this.pendingRebase = this._mergeTree.pendingSegments.splice(firstGroupNode);
        }
        const rebaseTo = this.getCollabWindow().currentSeq;
        if (rebaseTo !== this.lastNormalizationRefSeq) {
            this.emit("normalize", this);
            this._mergeTree.normalizeSegmentsOnRebase();
            this.lastNormalizationRefSeq = rebaseTo;
        }
        const opList = [];
        if (resetOp.type === MergeTreeDeltaType.GROUP) {
            if (Array.isArray(segmentGroup)) {
                assert(resetOp.ops.length === segmentGroup.length, 0x03a /* "Number of ops in 'resetOp' must match the number of segment groups provided." */);
                for (let i = 0; i < resetOp.ops.length; i++) {
                    opList.push(...this.resetPendingDeltaToOps(resetOp.ops[i], segmentGroup[i]));
                }
            }
            else {
                // A group op containing a single op will pass a direct reference to 'segmentGroup'
                // rather than an array of segment groups.  (See 'peekPendingSegmentGroups()')
                assert(resetOp.ops.length === 1, 0x03b /* "Number of ops in 'resetOp' must match the number of segment groups provided." */);
                opList.push(...this.resetPendingDeltaToOps(resetOp.ops[0], segmentGroup));
            }
        }
        else {
            assert(resetOp.type !== MergeTreeDeltaType.GROUP, 0x03c /* "Reset op has 'group' delta type!" */);
            assert(!Array.isArray(segmentGroup), 0x03d /* "segmentGroup is array rather than singleton!" */);
            opList.push(...this.resetPendingDeltaToOps(resetOp, segmentGroup));
        }
        // eslint-disable-next-line import/no-deprecated
        return opList.length === 1 ? opList[0] : createGroupOp(...opList);
    }
    // eslint-disable-next-line import/no-deprecated
    createTextHelper() {
        return new MergeTreeTextHelper(this._mergeTree);
    }
    summarize(runtime, handle, serializer, catchUpMsgs) {
        const deltaManager = toDeltaManagerInternal(runtime.deltaManager);
        const minSeq = deltaManager.minimumSequenceNumber;
        // Catch up to latest MSN, if we have not had a chance to do it.
        // Required for case where FluidDataStoreRuntime.attachChannel()
        // generates summary right after loading data store.
        this.updateSeqNumbers(minSeq, deltaManager.lastSequenceNumber);
        // One of the summaries (from SPO) I observed to have chunk.chunkSequenceNumber > minSeq!
        // Not sure why - need to catch it sooner
        assert(this.getCollabWindow().minSeq === minSeq, 0x03e /* "minSeq mismatch between collab window and delta manager!" */);
        // Must continue to support legacy
        //       (See https://github.com/microsoft/FluidFramework/issues/84)
        if (this._mergeTree.options?.newMergeTreeSnapshotFormat === true) {
            assert(catchUpMsgs === undefined || catchUpMsgs.length === 0, 0x03f /* "New format should not emit catchup ops" */);
            const snap = new SnapshotV1(this._mergeTree, this.logger, (id) => this.getLongClientId(id));
            snap.extractSync();
            return snap.emit(serializer, handle);
        }
        else {
            const snap = new SnapshotLegacy(this._mergeTree, this.logger);
            snap.extractSync();
            return snap.emit(catchUpMsgs, serializer, handle);
        }
    }
    async load(runtime, storage, serializer) {
        const loader = new SnapshotLoader(runtime, this, this._mergeTree, this.logger, serializer);
        return loader.initialize(storage);
    }
    getLocalSequenceNumber() {
        const segWindow = this.getCollabWindow();
        return segWindow.collaborating ? UnassignedSequenceNumber : UniversalSequenceNumber;
    }
    // eslint-disable-next-line import/no-deprecated
    localTransaction(groupOp) {
        for (const op of groupOp.ops) {
            const opArgs = {
                op,
                groupOp,
            };
            switch (op.type) {
                case MergeTreeDeltaType.INSERT: {
                    this.applyInsertOp(opArgs);
                    break;
                }
                case MergeTreeDeltaType.ANNOTATE: {
                    this.applyAnnotateRangeOp(opArgs);
                    break;
                }
                case MergeTreeDeltaType.REMOVE: {
                    this.applyRemoveRangeOp(opArgs);
                    break;
                }
                case MergeTreeDeltaType.OBLITERATE_SIDED:
                case MergeTreeDeltaType.OBLITERATE: {
                    this.applyObliterateRangeOp(opArgs);
                    break;
                }
                default: {
                    break;
                }
            }
        }
    }
    updateMinSeq(minSeq) {
        this._mergeTree.setMinSeq(minSeq);
    }
    getContainingSegment(pos, sequenceArgs, localSeq) {
        const { referenceSequenceNumber, clientId } = this.getClientSequenceArgsForMessage(sequenceArgs);
        return this._mergeTree.getContainingSegment(pos, referenceSequenceNumber, clientId, localSeq);
    }
    getPropertiesAtPosition(pos) {
        let propertiesAtPosition;
        const segoff = this.getContainingSegment(pos);
        const seg = segoff.segment;
        if (seg) {
            propertiesAtPosition = seg.properties;
        }
        return propertiesAtPosition;
    }
    getRangeExtentsOfPosition(pos) {
        let posStart;
        let posAfterEnd;
        const segoff = this.getContainingSegment(pos);
        const seg = segoff.segment;
        if (seg) {
            posStart = this.getPosition(seg);
            posAfterEnd = posStart + seg.cachedLength;
        }
        return { posStart, posAfterEnd };
    }
    getCurrentSeq() {
        return this.getCollabWindow().currentSeq;
    }
    getClientId() {
        return this.getCollabWindow().clientId;
    }
    getLength() {
        return this._mergeTree.length ?? 0;
    }
    startOrUpdateCollaboration(longClientId, minSeq = 0, currentSeq = 0) {
        // we should always have a client id if we are collaborating
        // if the client id is undefined we are likely bound to a detached
        // container, so we should keep going in local mode. once
        // the container attaches this will be called again on connect with the
        // client id
        if (longClientId !== undefined) {
            if (this.longClientId === undefined) {
                this.longClientId = longClientId;
                this.addLongClientId(this.longClientId);
                this._mergeTree.startCollaboration(this.getShortClientId(this.longClientId), minSeq, currentSeq);
            }
            else {
                const oldClientId = this.longClientId;
                const oldData = this.clientNameToIds.get(oldClientId).data;
                this.longClientId = longClientId;
                this.clientNameToIds.put(longClientId, oldData);
                this.shortClientIdMap[oldData] = longClientId;
            }
        }
    }
    /**
     * Searches a string for the nearest marker in either direction to a given start position.
     * The search will include the start position, so markers at the start position are valid
     * results of the search. Makes use of block-accelerated search functions for log(n) complexity.
     *
     * @param startPos - Position at which to start the search
     * @param markerLabel - Label of the marker to search for
     * @param forwards - Whether the desired marker comes before (false) or after (true) `startPos`
     */
    searchForMarker(startPos, markerLabel, forwards = true) {
        const clientId = this.getClientId();
        return this._mergeTree.searchForMarker(startPos, clientId, markerLabel, forwards);
    }
}
//# sourceMappingURL=client.js.map